/*
 * Copyright (c) 2004-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 123 $
 * $LastChangedDate: 2007-02-23 15:43:45 +0100 (Fri, 23 Feb 2007) $
 * 
 */

package com.codedroids.j;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Set;

public final class MapUtils {

	/**
	 * Populate a map from collections of keys and values
	 * @param map
	 * @param keys
	 * @param values
	 * @return
	 */
	public static Map populateMap(Map map, Collection keys, Collection values) 
	{
		Iterator k = keys.iterator();
		Iterator v = values.iterator();
		while(k.hasNext())
			map.put(k.next(), v.next()); // throws if we're short on values, but that's fine...
		return map;
	}
	
	/**
	 * Mimics the CSS prop format (foo:bar;and:one;...)
	 * @param map
	 * @param source
	 * @return
	 */
	public static Map populateMapFromString(Map map, String source)
	{
		return populateMapFromString(map, source, ";", ":");
	}
	
	/**
	 * Allows 2 dimensional format (foo:bar;and:one;...). There are no special escape char.
	 * @param map
	 * @param source
	 * @param majorSep major separator (e.g. ";" in above example)
	 * @param minorSep minor separator (e.g. ":" in above example)
	 * @return
	 */
	public static Map populateMapFromString(Map map, String source, String majorSep, String minorSep)
	{
		String[] a = source.split(majorSep);
		for(int i = 0; i<a.length; i++) {
			String[] b = a[i].split(minorSep);
			if(b.length == 1)
				map.put(b[0], null);
			else
				map.put(b[0], b[1]);
		}
		return map;
	}

	/**
	 * Allows 3 dimensional format (foo:bar;and:one,two,three;empty;...). There are no special escape char.<br>
	 * Values are returned as String arrays (or null if minorSep was found in a part string - as for "empty" in example)
	 * @param map
	 * @param source
	 * @param majorSep major separator (e.g. ";" in above example)
	 * @param minorSep minor separator (e.g. ":" in above example)
	 * @param valueSep value separator (e.g. "," in above example)
	 * @return
	 */
	public static Map populateMapFromString(Map map, String source, String majorSep, String minorSep, String valueSep)
	{
		String[] a = source.split(majorSep);
		for(int i = 0; i<a.length; i++) {
			String[] b = a[i].split(minorSep);
			if(b.length == 1)
				map.put(b[0], null);
			else
				map.put(b[0], b[1].split(valueSep));
		}
		return map;
	}
	
	public static HashMap getHashMap(Collection keys, Collection values) 
	{
		HashMap map = new HashMap(keys.size());
		return (HashMap)populateMap(map, keys, values);
	}
	
	public static String getAsString(Map map, Object key, String def) {
		Object o = map.get(key);
		if(o==null)
			return def;
		else
			return o.toString();
	}
	
	public static String getString(Map map, Object key, String def) {
		Object o = map.get(key);
		if(o==null)
			return def;
		else
			return (String)o;
	}

	public static List listFromMapOfMaps(Map map, String key, boolean skipNulls) {
		List list = new ArrayList(map.size());
		Iterator i = map.values().iterator();
		while(i.hasNext()) {
			Map m = (Map)i.next();
			Object o = m.get(key);
			if(o!=null || skipNulls)
				list.add(o);
		}
		return list;
	}
	
	public static String concat(Map map, String separator, String lastSeparator) {
		if(map == null)
			return null;
		StringBuffer b = new StringBuffer();
		if(separator==null)
			separator = "";
		if(lastSeparator==null)
			lastSeparator = separator;
		int counter = 0, last = map.size()-1;
		Iterator i = map.values().iterator();
		while(i.hasNext()) {
			if(counter > 0) {
				if(counter == last)
					b.append(lastSeparator);
				else
					b.append(separator);
			}
			b.append(i.next().toString());
			counter++;
		}
		return b.toString();
	}
	
	public static String prettyPrint(Map map) {
		return prettyPrint(map, null);
	}
	public static String prettyPrint(Map map, Set filter) {
		Set set = map.entrySet();
		return prettyPrint(set, filter);
	}
	/**
	 * Collection must contain Map.Entry's or sub classes thereof
	 * @param collection
	 * @param filter
	 * @return
	 */
	public static String prettyPrint(Collection collection) {
		return prettyPrint(collection, null);
	}
	public static String prettyPrint(Collection collection, Set filter) {
		StringBuffer buf = new StringBuffer();
		
		int maxWidth = 0;
		
		Iterator entries = collection.iterator();
		while(entries.hasNext()) {
			Map.Entry entry = (Map.Entry)entries.next();
			if(filter == null || filter.contains(entry.getKey())) {
				int labelLen = ((String)entry.getKey()).length();
				if(labelLen > maxWidth)
					maxWidth = labelLen;
			}
		}
		
		// left-justify padded labels
		entries = collection.iterator();
		while(entries.hasNext()) {
			Map.Entry entry = (Map.Entry)entries.next();
			if(filter == null || filter.contains(entry.getKey())) {
				String label = (String)entry.getKey();
				Object obj = entry.getValue();
				if(obj instanceof Object[]) {
					Object list[] = (Object[])obj;
					for(int i = 0; i < list.length; i++) {
						if(i==0)
							buf.append(StringUtils.padLeft(label, maxWidth, '.') + ".: " + list[i] + "\n");
						else
							buf.append(StringUtils.padLeft("", maxWidth, ' ') + " : " + list[i] + "\n");
					}
				} else {
					buf.append(StringUtils.padLeft(label, maxWidth, '.') + ".: " + obj + "\n");
				}
				
			}
		}
		return buf.toString();
	}
}
