/*
 * Copyright (c) 2006-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 182 $
 * $LastChangedDate: 2007-07-05 11:13:28 +0200 (Thu, 05 Jul 2007) $
 * 
 */

package com.codedroids.oc;

import org.opencms.file.CmsFile;
import org.opencms.file.CmsResource;
import org.opencms.jsp.CmsJspBean;
import org.opencms.main.CmsException;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Properties;
import java.util.Locale;
import java.util.Map;
import java.util.HashMap;

/**
 * Subclasses Properties to make a cacheable version.
 * 
 * <p>When looking up the cached object by means of the convenience method 
 * <code>getCachedProperties</code>, the hierarchy defined by the <code>default</code> field is checked
 * such that all parents to the current object are guarenteed to be fresh.
 * 
 * <p>Currently it will throw a CmsException 
 * 
 */
public class OCCachedProperties extends Properties implements I_OCCacheElement {

    private static final long serialVersionUID = 1L;
    
	// build list with highest priority in first position, like "foobar_en_US.properties", "foobar_en.properties", "foobar.properties"
	private static String[] getLocalizedPaths(Locale locale, String path) {
		String loc = locale.toString();
		String[] parts = loc.split("_");
		int len = parts.length;
		String[] paths = new String[len+1];
		paths[len] = path;
		StringBuffer partloc = new StringBuffer();
		for(int i=0; i<len; i++) {
			partloc.append(parts[i]);
			paths[len-i-1] = PathUtils.getLocalizedUriByFile(path, partloc.toString());
			partloc.append("_");
		}
		return paths;
	}
	
	/**
	 * Convenience for loading/caching properties, with or without locale-variants (locale is
	 * extracted from the cms context).
	 * 
	 * <p>It will silently ignore if any of the files are missing.
	 * 
	 * @param cache An initialized OCCache.
	 * @param cms An initialized Cms bean.
	 * @param path The VFS path (without site root).
	 * @param l10n If true languages variants based on the locale for request are loaded as well as the file given by the path.
	 * @return A <code>Properties</code> object (possibly the most language specific in case that <code>l10n</code> is <code>true</code>) or <code>null</code>.
	 * @throws CmsException.
	 */
	public static Properties cacheProperties(OCCache cache, CmsJspBean cms, String path, boolean l10n) throws CmsException {
		return cacheProperties(cache, cms, path, (l10n ? cms.getRequestContext().getLocale() : null));
	}
	
	/**
	 * Convenience for looking up (and possible loading/caching) properties, with or without locale-variants.
	 * 
	 * @param cache An initialized OCCache
	 * @param cms An initialized Cms bean
	 * @param path The VFS path (without site root)
	 * @param locale If given locale variants are cached as well, otherwise only the file spec'ed by path is cached.
	 * @return A <code>Properties</code> object (possibly the most language specific in case that <code>locale</code> is given) or <code>null</code>
	 * @throws CmsException
	 */
	public static Properties cacheProperties(OCCache cache, CmsJspBean cms, String path, Locale locale) throws CmsException {
		String[] paths = null;
		if(locale != null) {
			paths = getLocalizedPaths(locale, path);			
		} else {
			paths = new String[] {path};
		}
		Properties el = cacheProperties(cache, cms, paths, Boolean.FALSE);
		return el;
	}
	
	/**
	 * Loads the listed propery files such that the first in the array will have the highest priority and the
	 * next will become the defaults for the first and so forth. 
	 * 
	 * @param cache An initialized OCCache
	 * @param cms An initialized Cms bean
	 * @param paths List of VFS path (without site root)
	 * @param required If true an exception is thrown in case one of the files are missing
	 * @return The top-most (highest priority) Properties bundle of the ones listed in <code>paths</code> or <code>null</code>
	 * @throws CmsException
	 */
	public static Properties cacheProperties(OCCache cache, CmsJspBean cms, String[] paths, Boolean required) throws CmsException {
		OCCachedProperties prev = null, el = null;
		for(int i=paths.length-1; i>=0; i--) {
			String path = paths[i];
			HashMap context = new HashMap();
			context.put("path", path);
			context.put("parent", prev);
			context.put("required", required); 
			el = (OCCachedProperties)cache.cacheElement(cms, cms.getRequestContext().addSiteRoot(path), OCCachedProperties.class, context);
			if(el != null)
				prev = el;
		}
		return el;
	}
	
	/**
	 * Convenience for looking up (and possible loading/caching) properties, with or without locale-variants (locale is
	 * extracted from the cms context).
	 * 
	 * <p>It will silently ignore if any of the files are missing
	 * 
	 * @param cache An initialized OCCache
	 * @param cms An initialized Cms bean
	 * @param path The VFS path (without site root)
	 * @param l10n If true languages variants based on the locale for request are loaded as well as the file given by the path
	 * @return A <code>Properties</code> object (possibly the most language specific in case that <code>l10n</code> is <code>true</code>) or <code>null</code>
	 * @throws CmsException
	 */
	public static Properties getCachedProperties(OCCache cache, CmsJspBean cms, String path, boolean l10n) throws CmsException {
		return getCachedProperties(cache, cms, path, (l10n ? cms.getRequestContext().getLocale() : null));
	}
	
	/**
	 * Convenience for looking up (and possible loading/caching) properties, with or without locale-variants.
	 * 
	 * @param cache An initialized OCCache
	 * @param cms An initialized Cms bean
	 * @param path The VFS path (without site root)
	 * @param locale If given locale variants are cached as well, otherwise only the file spec'ed by path is cached.
	 * @return A <code>Properties</code> object (possibly the most language specific in case that <code>locale</code> is given) or <code>null</code>
	 * @throws CmsException
	 */
	public static Properties getCachedProperties(OCCache cache, CmsJspBean cms, String path, Locale locale) throws CmsException {
		String[] paths = null;
		if(locale != null) {
			paths = getLocalizedPaths(locale, path);			
		} else {
			paths = new String[] {path};
		}
		return (Properties)getCachedProperties(cache, cms, paths, Boolean.FALSE);

	}
	
	/**
	 * Looks up cached property files or caches them if not already cached, it hooks them up so the first in the list has the
	 * highest priority and the next becomes the defaults for the first and so forth. 
	 * 
	 * @param cache An initialized OCCache
	 * @param cms An initialized Cms bean
	 * @param paths List of VFS path (without site root)
	 * @param required If true an exception is thrown in case one of the files are missing
	 * @return The top-most (highest priority) Properties bundle of the ones listed in <code>paths</code> or <code>null</code>
	 * @throws CmsException
	 */
	public static Properties getCachedProperties(OCCache cache, CmsJspBean cms, String[] paths, Boolean required) throws CmsException {
		OCCachedProperties prev = null, el = null;
		for(int i=paths.length-1; i>=0; i--) {
			String path = paths[i];
			HashMap context = new HashMap();
			context.put("path", path);
			context.put("parent", prev);
			context.put("required", required);
			el = (OCCachedProperties)cache.getCachedElement(cms, cms.getRequestContext().addSiteRoot(path), OCCachedProperties.class, context);
			if(el != null)
				prev = el;
		}
		return el;
	}
	
	//
	// INSTANCE
	//
	
	private boolean initialized;
	private String path;
	private boolean online;
	private Boolean required;
	private long expires;
	private long lastmod;
	
	/**
	 * Context must be a Map with following key-value pairs.
	 * <ul>
	 * <li><code>path</code> the path of the file
	 * <li><code>parent</code> the default/parent object for this properties file
	 * <li><code>required</code> a <code>Boolean</code> to indicate if the path must exist - if <code>false</code> an unknown path
	 * does not cause an exception, instead the method returns <code>true</true>, but the object is not considered initialized.<br>
	 * <code>isStale()</code> then reports accordingly so the end result is that the cache ends up checking for the missing
	 * files over and over.
	 * </ul>
	 */
	public boolean init(CmsJspBean cms, boolean online, String key, Object context) throws CmsException {
		this.online = online;
		this.path = (String)((Map)context).get("path");
		this.defaults = (Properties)((Map)context).get("parent");
		this.required = (Boolean)((Map)context).get("required");
		return init(cms);
	}
	
	private boolean init(CmsJspBean cms) throws CmsException {
		clear();
		this.initialized = false;
		CmsResource res = null;
		try {
			res = cms.getCmsObject().readResource(path);
		} catch (CmsException e) {
			if(required.booleanValue())
				throw e;
			else
				return true;
		}
		if(res != null) {
			CmsFile f = CmsFile.upgrade(res,cms.getCmsObject());
			try {
				ByteArrayInputStream is = new ByteArrayInputStream(f.getContents());
				load(is);
				this.initialized = true;
				this.expires = res.getDateExpired();
				this.lastmod = res.getDateLastModified();
			} catch (IOException e) {
				throw new RuntimeException("Problems encountered while loading \""+ this.path + "\" as a property resource bundle.", e);
			}
			return true;
		}
		return false;
	}

	public boolean refresh(CmsJspBean cms) throws CmsException {
		return init(cms);
	}

	/**
	 * Reports if the object is stale and should be refreshed or not.
	 * 
	 * <p>For the normal case the following logic apply:
	 * <ul>
	 * <li>If cached in the online part it will always return false unless the files expire date has been reached.
	 * <li>If cached in the offline part it will check last modified date to see if the files has been changed since it was cached.
	 * </ul>
	 * 
	 * <p>In case the object is not fully initialized (which happens if the load context told <code>init()</code> that it should not
	 * throw on missing files, and the file then was missing), then the logic is:
	 * <ul>
	 * <li>If cached in the online part, this means that we can consider the object to be fresh as the file will not be readable until it 
	 * is published to the online project, and that operation will flush the cache anyway and thus cause it to be reloaded if possible.
	 * <li>If cached in the offline part it reports the object to be stale if the file now exists.
	 * </ul>
	 */
	public boolean isStale(CmsJspBean cms) throws CmsException {
		if(online) {
			if(initialized) {
				// Online can only become stale because of the expire date (and after publish, but OCCache handles that case)
				return (this.expires <= System.currentTimeMillis());
			} else {
				// it was not found during init() and it certainly won't be in offline until the next publish, so this will
				// never get stale unless clear by a publish anyway.
				return false;
			}
		} else {
			if(initialized) {
				CmsResource res = null;
				try {
					res = cms.getCmsObject().readResource(path);
				} catch (CmsException e) {
					// This can happen if it has become unreadable after it has been cached, refresh will run into the same problem
					// and either throw or make it uninitialized as per initial context setup
					return true; 
				}
				return !(res.getDateLastModified() <= this.lastmod);
			} else {
				// report it stale if the files exists now
				return cms.getCmsObject().existsResource(path);
			}
		}
	}

}
