/*
 * Copyright (c) 2003-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 117 $
 * $LastChangedDate: 2007-02-20 11:24:48 +0100 (Tue, 20 Feb 2007) $
 * 
 */
package com.codedroids.oc;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.*;

import org.opencms.file.*;
import org.opencms.i18n.CmsMessages;
import org.opencms.i18n.CmsMessageException;
import org.opencms.jsp.CmsJspActionElement;
import org.opencms.main.*;

/**
 * Mimics the CmsMessage class, but provides lookup in multiple prioritized 'bundles'<br>
 * Handles various aspects of dealing with property files.
 * <ol>
 * <li>Files are accessed through the VFS, i.e. the files need not be "mirrored" to specific locations in the real filesystem
 * <li>Implements caching a la FlexCache, i.e. online is cached until a publish takes place, offline is generally not cached,
 *     but the modification date is checked and furthermore a minimum refresh period can be given (i.e. only load every 3600 seconds).
 * <li>More files can be combined, i.e. foobar.setup and foobar_en.setup - convenience methods for dealing with this scenario
 *     prioritizes the files such that a lookup will first check the foobar_en variant if locale is en and if not found there it will
 *     lookup the key in the foobar.setup. This makes it easier to manage setup information that consist of both locale dependent and
 *     independent key-value pairs. With manual configuration this can be extended to add more 'layers', e.g. have both a foobar_factory
 *     and a foobar_custom. 
 * </ol>
 * 
 * Wraps get methods to gracefully handle exceptions
 */
public class Setup {
	//private Locale m_locale;
	private LinkedList bundles = new LinkedList();
	
	public static final int HIGHEST_PRIORITY = 0;
	public static final int LOWEST_PRIORITY = 1;
	public static final String SETUP_EXT = ".setup";
	
	private static Map formsCache;
	private static String CACHE_TIMESTAMP = "timestamp";
	private static String CACHE_INSTANCE = "instance";
    
	static {
		formsCache = new HashMap();
	}
	
	/**
	 * Caches the instances of Property Resource Bundles rather than the setup itself (since
	 * the same PRB can be used in more than one setup instance - typically if there is a
	 * locale independent PRB)
	 * 
	 * @param cmso
	 * @param setupfile
	 * @return
	 */
	public static synchronized PropertyResourceBundle getPRBInstance(CmsObject cmso, String setupfile)
	{
		Map m = (Map)formsCache.get(setupfile);
		PropertyResourceBundle prb = null;
		if(m != null) {
			long timestamp = ((Date)m.get(CACHE_TIMESTAMP)).getTime();
			try {
				CmsResource f = cmso.readResource(setupfile);
				if(f != null && f.getDateLastModified() <= timestamp)
					prb = (PropertyResourceBundle)m.get(CACHE_INSTANCE);
			} catch (CmsException e) {
				return null; 
			}
		}
		if(prb == null) {
			prb = getPRBFromVFS(cmso, setupfile);
			if(prb == null)
				return null; // didn't load, don't cache then
			m = new HashMap();
			m.put(CACHE_INSTANCE, prb);
			m.put(CACHE_TIMESTAMP, new Date());
			formsCache.put(setupfile, m);
			//A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[Setup] cached property resource bundle = " + setupfile);
		}
		return prb;
	}
	
	/**
	 * Convenience for finding a .properties file from the VFS
	 * 
	 * @param cmso CmsObject for the JSP calling this
	 * @param file filename and path relative to the root of VFS
	 */
	public static PropertyResourceBundle getPRBFromVFS(CmsObject cmso, String file)
	{
		try {
			CmsFile f = cmso.readFile(file);
			ByteArrayInputStream i = new ByteArrayInputStream(f.getContents());
			return new PropertyResourceBundle(i);  
		} catch (IOException e) {
			return null;
		} catch (CmsException e) {
			return null;
		}
	}
	
	/**
	 * Convenience for building a setup from a list of VFS directories.<br>
	 * It takes each path in the comma-separated <code>dirs</code> parameter,
	 * appends the optional <code>subdir</code> before appending the <code>filename</code>
	 * 
	 * @param cmso CmsObject for the JSP calling this
	 * @param dirs One or more comma-separated absolute paths, first gets highest priority
	 * @param subdir optional directy to be appended to each directory (i.e. "en")
	 * @param filename name of file to look for in the directories
	 */
	public static Setup getSetup(CmsObject cmso, String dirs, String subdir, String filename)
	{
		String[] paths = dirs.split("\\s*,\\s*");
		if(subdir==null)
			subdir="";
		Setup retval = new Setup((ResourceBundle)null);
		for(int i=0; i<paths.length; i++) {
			String s = paths[i] + "/" + subdir + "/" + filename;
			s = s.replaceAll("/+", "/");
			PropertyResourceBundle tmp = getPRBInstance(cmso, s);
			if(tmp != null) {
				retval.addWithPriority(tmp, LOWEST_PRIORITY);
//				if(i==0)
//					A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[Setup] added \"" + s + "\" (highest priority)");
//				else
//					A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[Setup] added \"" + s + "\"");
			}
		}
		return retval;
	}

	/**
	 * Setup for given file, and possibly a language variant
	 * @param cmso
	 * @param file
	 * @param language
	 * @return
	 */
	public static Setup getSetupFile(CmsObject cmso, String file, String locale)
	{
		String[] tryFiles = new String[2];
		int ndx;
		if((ndx = file.lastIndexOf(SETUP_EXT)) >= 0)
			file = file.substring(0, ndx);
		tryFiles[0] = file + SETUP_EXT;
		if(locale != null)
			tryFiles[1] = file + "_" + locale + SETUP_EXT;
		return getSetup(cmso, tryFiles, HIGHEST_PRIORITY);
	}
	
	/**
	 * Attempt to load property files from the given paths starting from the first in paths array
	 * Add them to setup with the given priority
	 */
	public static Setup getSetup(CmsObject cmso, String[] paths, int priority)
	{
		Setup retval = new Setup((ResourceBundle)null);
//		boolean first = true;
		for(int i=0; i<paths.length; i++) {
			String s = paths[i];
			PropertyResourceBundle tmp = getPRBInstance(cmso, s);
			if(tmp != null) {
				retval.addWithPriority(tmp, priority);
//				if(first)
//					A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[Setup] added \"" + s + "\" (" + (priority == LOWEST_PRIORITY ? "highest" : "lowest") + " priority)");
//				else
//					A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[Setup] added \"" + s + "\"");
//				first = false;
			}
		}
		return retval;
	}
	
	/**
     * Construct a Setup with one bundle, more can be added subsequently.<br>
     * Generally passing <code>null</code> for <code>bundle</code> in these methods are allowed
     * (i.e. ignored), the idea being that the <code>key()</code> method will 'detect' this
     * subsequently by returning the warning, thus preventing exceptions.
     * 
     * @param bundle this will have top priority in lookup (may be <code>null</code>)
     */
	public Setup(ResourceBundle bundle)
	{
		if(bundle != null)
			bundles.add(bundle);
	}
	
	/**
	 * Construct a Setup with one message bundle, more can be added subsequently
	 * 
	 * @param bundle this will have top priority in lookup
	 */
	public Setup(CmsMessages bundle)
	{
		if(bundle != null)
			bundles.add(bundle);
	}
	
	/**
	 * Construct a Setup with a list of bundles, more can be added subsequently
	 * 
	 * @param first element in list will get top priority (may be empty but not null)
	 */
	public Setup(Collection list)
	{
		bundles.addAll(list);
	}

	/**
	 * Add a <code>bundle</code> with given <code>priority</code>
	 * 
	 * @param bundle 
	 * @param priority <code>Setup.HIGHEST</code> or <code>Setup.LOWEST</code>
	 */
	public void addWithPriority(ResourceBundle bundle, int priority)
    {
		if(bundle != null) {
    		if(priority == HIGHEST_PRIORITY)
    			bundles.addFirst(bundle);
    		else
    		   	bundles.addLast(bundle);
		}	
    }
    
	/**
	 * Add a <code>bundle</code> with given <code>priority</code>
	 * 
	 * @param bundle 
	 * @param priority <code>Setup.HIGHEST</code> or <code>Setup.LOWEST</code>
	 */
	public void addWithPriority(CmsMessages bundle, int priority)
	{
		if(bundle != null) {
			if(priority == HIGHEST_PRIORITY)
				bundles.addFirst(bundle);
			else
				bundles.addLast(bundle);
		}
	}    
    
	/**
     * Checks if the setup was properly initialized.
     * 
	 * @return <code>true</code> if setup was initialized, <code>false</code> otherwise
	 */
	public boolean isInitialized() {
		return (bundles != null && !bundles.isEmpty());
	}
        
	/**
	 * Gets the localized resource string for a given message key.<p>
	 * 
	 * If the key was not found in the bundle, the return value is
	 * <code>"??? " + keyName + " ???"</code>. This will also be returned 
	 * if the bundle was not properly initialized first.
	 * 
	 * @param keyName the key for the desired string 
	 * @return the resource string for the given key 
	 */
	public String key( String keyName ) {   
		try {            
			return getString( keyName );
		} catch (Exception e) {
			// not found, return warning
		}
		return "??? " + keyName + " ???";
	}    
 
	/**
	 * Gets the localized resource string for a given message key or return the default.<p>
	 * 
	 * If the key was not found in the bundle, the return value is
	 * <code>null</code>. This is more useful for other things than strings to display
	 * 
	 * @param keyName the key for the desired string 
	 * @return the resource string for the given key or null
	 */
	public String valueForKey( String keyName) {
		return valueForKey(keyName, null);
	}
	public String valueForKey( String keyName, String def ) {   
		try {            
			return getString( keyName );
		} catch (Exception e) {
			// not found, return warning
		}
		return def;
	}    
    
	
	/**
	 * Directly calls the getString(String) method of the wrapped ResourceBundle.<p>
	 * 
	 * If you use this this class on a template, you should consider using 
	 * the {@link #key(String)} method to get the value from the ResourceBundle because it
	 * handles the exception for you in a convenient way. 
	 * 
	 * @param keyName the key  
	 * @return the resource string for the given key
	 * @throws Exception in case the key is not found of the bundle is not initialized
	 */
	public String getString( String keyName ) throws Exception {              
		if (!isInitialized())
			throw new Exception("ResourceBundle not initialized: " + this.getClass().getName() + " - " + keyName);
		Iterator li = bundles.iterator();
		while(li.hasNext()) {
			Object o = li.next();
			try {
				if(o instanceof CmsMessages)
					return ((CmsMessages)o).getString(keyName);
				else
					return ((ResourceBundle)o).getString(keyName);
			} catch (CmsMessageException e) {
				// ignore and try the next bundle
			}
		}
		throw new Exception("Key not found in any bundle in Setup: " + this.getClass().getName() + " - " + keyName);
	}
	
	// These mimics the similar methods on the CmsJspBean, only difference is that they all look for
	// the property in the setup file after looking in the resource properties and before deciding to use the
	// default value.
	
	public String property(CmsJspActionElement cms, String name) {
		return property(cms, name, null, null, false);
	}
	public String property(CmsJspActionElement cms, String name, String file) {
		return property(cms, name, file, null, false);
	}
	public String property(CmsJspActionElement cms, String name, String file, String defaultValue) {
		return property(cms, name, file, defaultValue, false);
	}
	public String property(CmsJspActionElement cms, String name, String file, String defaultValue, boolean escapeHtml) {
		String res = cms.property(name, file, null, escapeHtml);
		if(res == null)
			res = this.valueForKey(name);
		if(res == null)
			res = defaultValue;
		return res;
	}

}
