/*
 * Copyright (c) 2005-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 180 $
 * $LastChangedDate: 2007-06-20 11:22:10 +0200 (Wed, 20 Jun 2007) $
 * 
 */

package com.codedroids.records;

import java.lang.reflect.Method;
import java.text.ParseException;
import java.util.*;

/**
 * Basically a map (extending MapRecord), but it will try to use the standard getter and setter methods
 * on itself before forwarding the call to the underlying Map record storage. 
 * Covers the Map.get and Map.put methods to allow this and these methods also allow access to non-persisted
 * methods and key-value-pairs since they won't throw if the key is not a field.
 * 
 */
public class GenericRecord extends MapRecord {
	
	static final long serialVersionUID =1L;
	
	public GenericRecord() {
		super();
	}
	
	public GenericRecord(RecordFactory factory) {
		this();
		init(factory);
	}

	// The rest are constructors that instantiates an object from the database
	public GenericRecord(RecordFactory factory, String id) throws ParseException, RecordException {
		this();
		init(factory, factory.getIdentField().fromString(id));
	}
	
	public GenericRecord(RecordFactory factory, Object id) throws RecordException {
		this();
		init(factory, id);	
	}

	public void takeValueForField(RecordField field, Object value) {
		String methodName = "set" + field.getName().substring(0,1).toUpperCase() + field.getName().substring(1);
		try {
			Method m = this.getClass().getMethod(methodName, new Class[] { field.fieldClass() } );
			m.invoke(this, new Object[] { value });		
		} catch (NoSuchMethodException e) {
			super.takeValueForField(field, value);
		} catch (Exception e) {
			throw new IllegalArgumentException(getClass().getName() + " takeValueForField(): field=\""+field.getName()+"\" threw: "+e);
			//e.printStackTrace(); // Should probably throw the various exceptions...
		}
	}
	
	public Object valueForField(RecordField field) {
		String methodName = "get" + field.getName().substring(0,1).toUpperCase() + field.getName().substring(1);
		try {
			Method m = this.getClass().getMethod(methodName, null );
			return m.invoke(this, null);
		} catch (NoSuchMethodException e) {
			return super.valueForField(field);
		} catch (Exception e) {
			throw new IllegalArgumentException(getClass().getName() + " valueForField(): field=\""+field.getName()+"\" threw: "+e);
		}
	}
	
	/**
	 * Values can be given as strings which then will be parsed which in turn may throw an exception
	 * @param values
	 * @throws ParseException
	 */
	public void takeValuesFromMap(Map values) throws ParseException {
		if(values == null)
			return;
		Iterator i = values.entrySet().iterator();
		while(i.hasNext()) {
			Map.Entry entry = (Map.Entry)i.next();
			String key = (String)entry.getKey();
			RecordField field = factory.getField(key);
			if(field != null) {
				Object val = entry.getValue();
				if(val != null && val instanceof String)
					takeStringValueForField(field, (String)val);
			} // else just ignore it...
		}
	}
	
	// Cover for basic map-methods, enables freemarker and displaytag to work with these as well
	
	/**
	 * Don't wan't this to throw if the key is not a field (valueForKey() will do that)
	 * This is to support accessor methods defined on subclasses and non-persisted key-values
	 * in the map.
	 */
	public Object get(Object key) {
		if(key==null) {
			return super.get(null);
		} else {
			String name = key.toString();
			RecordField field = factory.getField(name);
			if(field != null) {
				return valueForField(field);
			} else {
				String methodName = "get" + name.substring(0,1).toUpperCase() + name.substring(1);
				try {
					Method m = this.getClass().getMethod(methodName, null );
					return m.invoke(this, null);
				} catch (NoSuchMethodException e) {
					return super.get(name);
				} catch (Exception e) {
					throw new IllegalArgumentException(getClass().getName() + " get(): key=\""+name+"\" threw: "+e);
				}
			}
		}
	}
	
	/**
	 * Ditto, setting a value on a non-persistent field can be done, but then it is, well, not persisted
	 */
	public Object put(Object key, Object value)  
	{
		if(key==null) {
			return super.put(null,value);
		} else {
			String name = key.toString();
			RecordField field = factory.getField(name);
			if(field != null) {
				Object old = valueForField(field);
				takeValueForField(field, value);
				return old;
			} else {
				String methodName = "set" + name.substring(0,1).toUpperCase() + name.substring(1);
				try {
					Method m = this.getClass().getMethod(methodName, new Class[] { value.getClass() } ); // TODO won't work if value is null...
					Object old = get(name);
					m.invoke(this, new Object[] { value });
					return old;
				} catch (NoSuchMethodException e) {
					return super.put(name, value);
				} catch (Exception e) {
					throw new IllegalArgumentException(getClass().getName() + " put(): key=\""+name+"\" threw: "+e);
					//e.printStackTrace(); // Should probably throw the various exceptions...
				}
			}
		}
	}

	/**
	 * This is really not accurate in the sense that any non-persistent fields is ignored, but
	 * it is better than just reporting the map fields
	 */
	public Set keySet() {
		return factory.getFields().keySet();
	}
	
	/**
	 * Ditto concern as for keySet()...
	 */
	public Collection values() {
		Set keys = keySet();
		ArrayList vals = new ArrayList(keys.size());
		Iterator fields = keys.iterator();
		while(fields.hasNext()) {
			vals.add(get(fields.next()));
		}
		return vals;
	}

}
